/****************************************************************************
 Title	:   Drehencoder Library
 Author:    Frank Klee
 Date:	    2014/12/04
 Target:    ATMega8

 DESCRIPTION
		Basisroutinen zum Abfragen eines Drehencoders mittels Polling.
		
		Bei Bedarf wird Timer0 fr das Polling verwendet.
		
		
 USAGE
       Siehe C Includefile encoder.h
       
*****************************************************************************/
#include <avr/io.h>
#include <avr/interrupt.h>
#include "encoder.h"

int8_t Drehgeber_Position;		//Zhlt beim drehen des Drehgebers auf und ab
int8_t Drehgeber_Max;			//Maximalwert des Drehgebers (wird von Hauptprogramm vorgegeben)
int8_t Drehgeber_Min;			//Minimalwert des Drehgebers (wird von Hauptprogramm vorgegeben)


//Registerdefinitionen der verschiedenen Prozessoren anpassen
#if  defined(__AVR_ATmega8__)
 #define Timer0_Interruptmask TIMSK
 #define Timer0_Control TCCR0

#elif defined(__AVR_ATmega48__) ||defined(__AVR_ATmega88__) ||\
   defined(__AVR_ATmega168__) || defined(__AVR_ATmega48P__) ||\
   defined(__AVR_ATmega88P__) || defined(__AVR_ATmega168P__) ||\
   defined(__AVR_ATmega328P__) || defined(__AVR_ATmega328__) ||\
   defined(__AVR_ATtiny24__) ||defined(__AVR_ATtiny44__) ||defined(__AVR_ATtiny84__)
 #define Timer0_Interruptmask TIMSK0
 #define Timer0_Control TCCR0B

#elif defined(__AVR_ATtiny2313__) ||(__AVR_ATtiny4313__)
 #define Timer0_Interruptmask TIMSK
 #define Timer0_Control TCCR0B

#else
 #error "no Timer definition for MCU available"
#endif
/* 
** constants/macros 
*/
#define DDR(x) (*(&x - 1))      /* address of data direction register of port x */
#define Port(x) (*(&x))          /* address of data register of port x */
#if defined(__AVR_ATmega64__) || defined(__AVR_ATmega128__)
    /* on ATmega64/128 PINF is on port 0x00 and not 0x60 */
    #define PIN(x) ( &PORTF==&(x) ? _SFR_IO8(0x00) : (*(&x - 2)) )
#else
	#define PIN(x) (*(&x - 2))    /* address of input register of port x          */
#endif


#define PHASE_A     (PIN(ENC_A_PORT) & 1<<ENC_A_PIN)
#define PHASE_B     (PIN(ENC_B_PORT) & 1<<ENC_B_PIN)
#define BUTTONPRESSED	(!(PIN(ENC_T_PORT) & (1<<ENC_T_PIN)))

#define BUTTON_DEBOUNCETIME_MS	30						// Zeit ber die entprellt werden soll in ms
#define BUTTON_PRESSEDLONG_MS	1000					// Zeitraum fr das Erkennen eines lang gedrckten Tasters

volatile int8_t enc_delta;          // -128 ... 127
static int8_t last;

volatile int16_t iButtonPressedCounter = 0;
volatile int16_t iButtonDebounceCycles = 0;
volatile int16_t iButtonPressedLongCycles = 0;

typedef enum EButtonState
{
	ButtonState_Unpressed,
	ButtonState_Pressed,
	ButtonState_Hold,
	ButtonState_Released	
}tEButtonState;	
volatile tEButtonState buttonState = ButtonState_Unpressed;
volatile tEButtonPressedState buttonPressed = ButtonPressed_Unpressed;

unsigned char TimerStart;
 
void encoder_init( void )
{
    DDR(ENC_A_PORT)    &= _BV(ENC_A_PIN);
    DDR(ENC_B_PORT)    &= _BV(ENC_B_PIN);
    DDR(ENC_T_PORT)    &= _BV(ENC_T_PIN);
	
	Port(ENC_A_PORT)   |= _BV(ENC_A_PIN);          // PullUp fr Drehimpulsgeber und Taster
    Port(ENC_B_PORT)   |= _BV(ENC_B_PIN);
    Port(ENC_T_PORT)   |= _BV(ENC_T_PIN);

#if (Use_T0_Interrupt) 
	/* init Timer0
	Timerberlauf alle 1ms
	Timer zhlt von 0 bis 255, dann OVR-Interrupt
	*/
	Timer0_Interruptmask |= (1 << TOIE0);				// Interrupt fr Timer0 Overflow freigeben
	Timer0_Control |= (1<<CS00) | (1<<CS01);		// use Prescaler /64
	TimerStart=256-(F_CPU/64/1000);
	TCNT0=TimerStart;
#endif
	
  int8_t new;
 
  new = 0;
  if( PHASE_A )
    new = 3;
  if( PHASE_B )
    new ^= 1;                   // convert gray to binary
  last = new;                   // power on state
  enc_delta = 0;

  iButtonDebounceCycles = BUTTON_DEBOUNCETIME_MS;		//Anzahl der ntigen Zyklen zum Entprellen berechnen
  iButtonPressedLongCycles = BUTTON_PRESSEDLONG_MS;	//Anzahl der ntigen Zyklen fr Lange Drcken berechnen

}

#if defined (Single_Step)
	int8_t encode_read( void )         // read single step encoders
	{
	  int8_t val;
 
	  cli();
	  val = enc_delta;
	  enc_delta = 0;
	  sei();
	  return val;                   // counts since last call
	}

#elif defined (Two_Step)
	int8_t encode_read( void )         // read two step encoders
	{
	  int8_t val;
 
	  cli();
	  val = enc_delta;
	  enc_delta = val & 1;
	  sei();
	  return val >> 1;                   // counts since last call
	}

#elif defined (Four_Step)
	int8_t encode_read( void )         // read four step encoders
	{
	  int8_t val;
 
	  cli();
	  val = enc_delta;
	  enc_delta = val & 3;
	  sei();
	  return val >> 2;
	}
#endif
 
//gibt die Drehgeberposition zurck
int8_t encoder_read(char Ueberlauf)
{
	Drehgeber_Position += encode_read();
	if (Drehgeber_Position > Drehgeber_Max)
	{
		if (Ueberlauf)
		{
			Drehgeber_Position = Drehgeber_Position - Drehgeber_Max + Drehgeber_Min -1 ;
		} 
		else
		{
			Drehgeber_Position = Drehgeber_Max;
		}
	}	
	if (Drehgeber_Position < Drehgeber_Min)
	{
		if (Ueberlauf)
		{
			Drehgeber_Position = Drehgeber_Position + Drehgeber_Max - Drehgeber_Min + 1;
		} 
		else
		{
			Drehgeber_Position = Drehgeber_Min;
		}
	}	
	return Drehgeber_Position;
}

//setzt die aktuelle Drehgeberposition
void encoder_write(int8_t EncoderPos)
{
	Drehgeber_Position=EncoderPos;
}
 
//setzt die Minimal- und Maximalwerte fr die absolute Drehgeberposition
void encoder_MinMax(int8_t EncoderMin,int8_t EncoderMax)
{
	Drehgeber_Min=EncoderMin;
	Drehgeber_Max=EncoderMax;
	if (Drehgeber_Position > Drehgeber_Max) Drehgeber_Position = Drehgeber_Max;
	if (Drehgeber_Position < Drehgeber_Min) Drehgeber_Position = Drehgeber_Min;
}

tEButtonPressedState encoder_GetButtonState()
{	
	tEButtonPressedState retVal = buttonPressed;
	buttonPressed = ButtonPressed_Unpressed;
	return retVal;
}

/*********************************************
 Abfrage des Drehencoders und des Tasters
 
 mu etwa alle 1ms aufgerufen werden
 
**********************************************/
void encoder_polling(void)
{
	int8_t new, diff;
	
	new = 0;
	if( PHASE_A )	new = 3;
	if( PHASE_B )   new ^= 1;                   // convert gray to binary
	diff = last - new;			                // difference last - new
	if( diff & 1 ){					            // bit 0 = value (1)
		last = new;				                // store new as next last
		enc_delta += (diff & 2) - 1;	        // bit 1 = direction (+/-)
	}
	

	//Button-Auswertung
	switch(buttonState)
	{
		case ButtonState_Unpressed:
			if (BUTTONPRESSED) buttonState = ButtonState_Pressed;
			break;
		
		case ButtonState_Pressed:
			buttonState = ButtonState_Hold;
			break;
		
		case ButtonState_Hold:
			iButtonPressedCounter ++;
			if (iButtonPressedCounter >= iButtonDebounceCycles && (!BUTTONPRESSED)) 
			{
				buttonState = ButtonState_Released;
				if (buttonPressed != ButtonPressed_Long) buttonPressed = ButtonPressed_Short;
			}				
			if (iButtonPressedCounter >= iButtonPressedLongCycles)
			{
				buttonPressed = ButtonPressed_Long;
			}
			break;
		
		case ButtonState_Released:
			iButtonPressedCounter = 0;
			buttonState = ButtonState_Unpressed;
			break;
	}
}

#if (Use_T0_Interrupt)
/*************************************************************

	Timer0 OVR
	
	wird ca. alle 1ms aufgerufen
	
	Fragt den Drehgeber incl. Taster ab
	
*************************************************************/

ISR(TIMER0_OVF_vect)
{
	TCNT0=TimerStart;
	encoder_polling();
}
#endif
